// code by Thierry Schneider
// adapted by jph

// checks existence of serial ports
int serial_detect(int top,int* arr) {
  char str[10];
  DCB dcb;
  int cnt=0;
  for (int c1=1;c1<=top && cnt<top;++c1) {
    sprintf(str,"COM%i",c1);
    if (BuildCommDCB(str,&dcb))
      arr[cnt++]=c1;
  }
  return cnt;
}

void serial_printf(DCB &dcb) {
  printf("DCBlength          %i\n",dcb.DCBlength);                 // DWORD 
  printf("BaudRate           %i\n",dcb.BaudRate);                  // DWORD 
  printf("fBinary            %i\n",dcb.fBinary);//  :1;            // DWORD 
  printf("fParity            %i\n",dcb.fParity);//  :1;            // DWORD 
  printf("fOutxCtsFlow       %i\n",dcb.fOutxCtsFlow);//  :1;       // DWORD 
  printf("fOutxDsrFlow       %i\n",dcb.fOutxDsrFlow);//  :1;       // DWORD 
  printf("fDtrControl        %i\n",dcb.fDtrControl);//  :2;        // DWORD 
  printf("fDsrSensitivity    %i\n",dcb.fDsrSensitivity);//  :1;    // DWORD 
  printf("fTXContinueOnXoff  %i\n",dcb.fTXContinueOnXoff);//  :1;  // DWORD 
  printf("fOutX              %i\n",dcb.fOutX);//  :1;              // DWORD 
  printf("fInX               %i\n",dcb.fInX);//  :1;               // DWORD 
  printf("fErrorChar         %i\n",dcb.fErrorChar);//  :1;         // DWORD 
  printf("fNull              %i\n",dcb.fNull);//  :1;              // DWORD 
  printf("fRtsControl        %i\n",dcb.fRtsControl);//  :2;        // DWORD 
  printf("fAbortOnError      %i\n",dcb.fAbortOnError);//  :1;      // DWORD 
  printf("fDummy2            %i\n",dcb.fDummy2);//  :17;           // DWORD 
  printf("wReserved          %i\n",dcb.wReserved);//               // WORD  
  printf("XonLim             %i\n",dcb.XonLim);                    // WORD  
  printf("XoffLim            %i\n",dcb.XoffLim);                   // WORD  
  printf("ByteSize           %i\n",dcb.ByteSize);                  // BYTE  
  printf("Parity             %i\n",dcb.Parity);                    // BYTE  
  printf("StopBits           %i\n",dcb.StopBits);                  // BYTE  
  printf("XonChar            %i\n",dcb.XonChar);                   // char  
  printf("XoffChar           %i\n",dcb.XoffChar);                  // char  
  printf("ErrorChar          %i\n",dcb.ErrorChar);                 // char  
  printf("EofChar            %i\n",dcb.EofChar);                   // char  
  printf("EvtChar            %i\n",dcb.EvtChar);                   // char  
  printf("wReserved1         %i\n",dcb.wReserved1);                // WORD  
}

// baud = 1200, 2400, 4800, 9600, 14400, 19200, 38400, 57600 and 115200 bit/s
void serial_init(DCB &dcb,int baud,int parity) {
  memset(&dcb,0,sizeof(dcb));
  dcb.DCBlength=sizeof(dcb);
  dcb.BaudRate=baud;
  dcb.fBinary=1; // If this member is TRUE, binary mode is enabled. Windows does not support nonbinary mode transfers, so this member must be TRUE.
  dcb.fParity=parity!=NOPARITY; // If this member is TRUE, parity checking is performed and errors are reported.  
  dcb.fOutxCtsFlow=0; // If this member is TRUE, the CTS (clear-to-send) signal is monitored for output flow control. If this member is TRUE and CTS is turned off, output is suspended until CTS is sent again.
  dcb.fOutxDsrFlow=0; // If this member is TRUE, the DSR (data-set-ready) signal is monitored for output flow control. If this member is TRUE and DSR is turned off, output is suspended until DSR is sent again.  
  // The DTR (data-terminal-ready) flow control. This member can be one of the following values.
  // DTR_CONTROL_DISABLE DTR_CONTROL_ENABLE DTR_CONTROL_HANDSHAKE
  dcb.fDtrControl=DTR_CONTROL_DISABLE;  
  dcb.fDsrSensitivity=0; // If this member is TRUE, the communications driver is sensitive to the state of the DSR signal. The driver ignores any bytes received, unless the DSR modem input line is high.
  dcb.fTXContinueOnXoff=0; // !!! If this member is TRUE, transmission continues after the input buffer has come within XoffLim bytes of being full and the driver has transmitted the XoffChar character to stop receiving bytes. If this member is FALSE, transmission does not continue until the input buffer is within XonLim bytes of being empty and the driver has transmitted the XonChar character to resume reception.
  dcb.fOutX=0; // Indicates whether XON/XOFF flow control is used during transmission. If this member is TRUE, transmission stops when the XoffChar character is received and starts again when the XonChar character is received.
  dcb.fInX=0; // Indicates whether XON/XOFF flow control is used during reception. If this member is TRUE, the XoffChar character is sent when the input buffer comes within XoffLim bytes of being full, and the XonChar character is sent when the input buffer comes within XonLim bytes of being empty.
  dcb.fErrorChar=0; // Indicates whether bytes received with parity errors are replaced with the character specified by the ErrorChar member. If this member is TRUE and the fParity member is TRUE, replacement occurs.
  dcb.fNull=0; // If this member is TRUE, null bytes are discarded when received.
  // The RTS (request-to-send) flow control. This member can be one of the following values.
  // RTS_CONTROL_DISABLE RTS_CONTROL_ENABLE RTS_CONTROL_HANDSHAKE RTS_CONTROL_TOGGLE
  dcb.fRtsControl=RTS_CONTROL_DISABLE;
  dcb.fAbortOnError=0; // If this member is TRUE, the driver terminates all read and write operations with an error status if an error occurs. The driver will not accept any further communications operations until the application has acknowledged the error by calling the ClearCommError function.
  // dcb.fDummy2; // Reserved; do not use.
  // dcb.wReserved; // Reserved; must be zero.
  dcb.XonLim=2; // The minimum number of bytes in use allowed in the input buffer before flow control is activated to allow transmission by the sender. This assumes that either XON/XOFF, RTS, or DTR input flow control is specified in the fInX, fRtsControl, or fDtrControl members.
  dcb.XoffLim=4; // The minimum number of free bytes allowed in the input buffer before flow control is activated to inhibit the sender. Note that the sender may transmit characters after the flow control signal has been activated, so this value should never be zero. This assumes that either XON/XOFF, RTS, or DTR input flow control is specified in the fInX, fRtsControl, or fDtrControl members. The maximum number of bytes in use allowed is calculated by subtracting this value from the size, in bytes, of the input buffer.
  dcb.ByteSize=8; // The number of bits in the bytes transmitted and received.
  // Parity is a method of detecting errors in transmission. When parity is used with a serial port, an extra data bit is sent with each data character, arranged so that the number of 1 bits in each character, including the parity bit, is always odd or always even. If a byte is received with the wrong number of 1s, then it must have been corrupted. However, an even number of errors can pass the parity check.
  // parity = NOPARITY==0, ODDPARITY==1, EVENPARITY==2  
  dcb.Parity=parity;
  // The number of stop bits to be used. This member can be one of the following values.
  // ONESTOPBIT==0, ONE5STOPBITS==1, TWOSTOPBITS==2
  dcb.StopBits=ONESTOPBIT;
  dcb.XonChar=0x13; // The value of the XON character for both transmission and reception.
  dcb.XoffChar=0x19; // The value of the XOFF character for both transmission and reception.
  dcb.ErrorChar=0; // The value of the character used to replace bytes received with a parity error.
  dcb.EofChar=0; // The value of the character used to signal the end of data.
  dcb.EvtChar=0; // The value of the character used to signal an event.
  dcb.wReserved1=0; // Reserved; must be zero.  
}

HANDLE serial_open(DCB &dcb,int port) {
  char str[10];
  sprintf(str,"COM%i",port);  
  HANDLE hnd=CreateFile(str,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,NULL,NULL); // opening serial port
  SetCommMask(hnd,0); // TODO for events!
  COMMTIMEOUTS tim;
  memset(&tim,0,sizeof(tim));
  tim.ReadIntervalTimeout=0;
  tim.ReadTotalTimeoutMultiplier=0;
  tim.ReadTotalTimeoutConstant=0;
  tim.WriteTotalTimeoutMultiplier=0;
  tim.WriteTotalTimeoutConstant=0;
  SetCommTimeouts(hnd,&tim);
  SetCommState(hnd,&dcb);
  return hnd;
}

int serial_close(const HANDLE& hnd) {
  return CloseHandle(hnd);
}

// returns number of bytes sent
// if serial port is closed -> returns 0
int serial_send(const HANDLE& hnd,char *buf,int len) {
  unsigned long num;
  WriteFile(hnd,buf,len,&num,0);
  return num;
}

// if the handle is invalid function returns 0
int serial_poll(const HANDLE& hnd) {
  COMSTAT sta;
  memset(&sta,0,sizeof(sta));
  unsigned long err;
  ClearCommError(hnd,&err,&sta);
  return sta.cbInQue; // The number of bytes received by the serial provider but not yet read by a ReadFile operation.
}

// len is "The maximum number of bytes to be read."
// if the handle is invalid function returns 0
// Important: if len is greater than the number of bytes received the function will stall!
int serial_read(const HANDLE& hnd,char *buf,int len) {
  unsigned long num;
  ReadFile(hnd,buf,len,&num,0);
  return num;
}
